module Views.LinePlot exposing (view, init, update, icon)
import Model exposing (Global, View(..))
import Msg exposing (Msg(..))
import Data.DataLoader exposing (Entry)

import Html exposing (Html, div, select, option, label)
import Html.Attributes
import Html.Events exposing (onInput)

import List.Extra exposing (unique)
import Dict

import Chart.Line as Line
import Scale.Color
import Chart.Symbol as Symbol
import Axis

import Config.LinePlotConfig as Local exposing (Model)
import Utils exposing (codeToName)

icon : Html Msg
icon =
    Html.i [ Html.Attributes.class "fa-solid fa-chart-line fa-xl" ] []

init : Model
init = { pointAnnotation = Nothing , vLineAnnotation = Nothing }

propagate : Local.Msg -> Msg
propagate msg =
    GotLinePlot msg

update : Local.Msg -> Global -> Model -> ( Global, Model, Cmd Msg )
update msg global model = 
    case msg of
        Local.CountryChange country ->
            ( { global | selectedCountry = Just country }, model, Cmd.none )
        Local.ConsulateChange consulateCountry ->
            ( { global | selectedConsulateCountry = Just consulateCountry }, model, Cmd.none )
        Local.Hint response ->
            ( global, { model
                | pointAnnotation =
                    response
                        |> Maybe.map (\hint -> ( hint, [( "fill","#fff" ), ( "stroke","#000" ), ( "stroke-width","1.5" )] ))
                , vLineAnnotation =
                    response
                        |> Maybe.map (\hint -> ( hint, [( "stroke", "#999999" )] ))
              }
            , Cmd.none
            )



type alias DataPoint =
    { groupLabel : String
    , x : Int
    , y : Int
    }



view : Global -> Model -> Html Msg
view global model = 
    let
        -- Default selected country
        country_default : String
        country_default =
            Maybe.withDefault "AT" global.selectedCountry

        -- Get a list of all countries from global.data
        countries : List String
        countries =
            List.map .state global.data
            |> unique

        -- Generate a list of country options for the select element
        country_option : String -> Html Msg
        country_option country =
            let
                selected : Bool
                selected =
                    case global.selectedCountry of
                        Nothing -> False
                        Just c -> country == c
            in
            option
            [ Html.Attributes.value country, Html.Attributes.selected selected ]
            [ Html.text (codeToName country) ]

        -- Get a list of all consulate countries from global.data
        consulate_countries : List String
        consulate_countries =
            let
                consulates : List (String, String)
                consulates = global.data
                    |> List.filter (\entry -> not (entry.totalIssuedMEV == Nothing || entry.totalVisasApplied == Nothing || entry.totalVisasIssued == Nothing))
                    |> List.map (\entry -> (entry.consulateCountry, entry.state))
                    |> List.filter (\(_, country) -> country == country_default)
            in
            List.map Tuple.first consulates
            |> unique
            |> List.filter (\x -> not (List.member x countries)) -- Filter out countries that are schengen states
            |> List.sortBy (\x -> (codeToName x))
            |> List.append [""] -- Add an empty string to show all consulates

        default_consulateCountry : String
        default_consulateCountry =
            let
                cc : String
                cc = Maybe.withDefault "" global.selectedConsulateCountry
            in
            if (List.member (cc) consulate_countries) then cc else ""

        -- Generate a list of consulate country options for the select element
        consulate_option : String -> Html Msg
        consulate_option country =
            option
            [ Html.Attributes.value country, Html.Attributes.selected (country == default_consulateCountry) ]
            [ Html.text (if not (country == "") then (codeToName country) else "-- Alle --") ]

        -- Filter global.data based on the selected visa type, consulate and country
        filterList : List Entry -> String -> String -> String -> List (Int, Int)
        filterList data country visatype consulate =
            data
            |> List.filter (\entry -> entry.state == country)
            -- If consulate is not empty, filter by consulate else show all
            |> List.filter (\entry -> if consulate == "" then True else entry.consulateCountry == consulate)
            |> List.map (\entry -> (entry.year, Maybe.withDefault 0 (
                case visatype of
                    "totalVisasApplied" -> entry.totalVisasApplied
                    "totalVisasIssued" -> entry.totalVisasIssued
                    _ -> Nothing
                )))
            -- Merge tuples of the same year (sum the visa count)
            |> List.foldl
                    (\(year, count) acc ->
                        Dict.update year
                            (\maybeCount ->
                                Just (Maybe.withDefault 0 maybeCount + count)
                            )
                            acc
                    )
                    Dict.empty
            |> Dict.toList
            |> List.map (\(year, count) -> (year, round count))
            -- |> List.filter (\(_, count) -> count > 0)


        -- Get a list of all years from global.data for the xaxis labeling of the selected country
        getYears : List Int
        getYears =
            let
                applied_years : List Int
                applied_years = List.map Tuple.first (filterList global.data country_default "totalVisasApplied" default_consulateCountry)

                issued_years : List Int
                issued_years = List.map Tuple.first (filterList global.data country_default "totalVisasIssued" default_consulateCountry)
            in
            List.append applied_years issued_years
                |> List.sort
                |> unique

        
        -- Wrap the result from filterList in List DataPoint
        wrapDataPoint : List (Int, Int) -> String -> List DataPoint
        wrapDataPoint data label =
            List.map (\(year, count) -> { groupLabel = label, x = year, y = count }) data

        
        filterVApplied : List DataPoint
        filterVApplied = wrapDataPoint (filterList global.data country_default "totalVisasApplied" default_consulateCountry) "applied"

        filterVIssued : List DataPoint
        filterVIssued = wrapDataPoint (filterList global.data country_default "totalVisasIssued" default_consulateCountry) "issued"
        

        -- Merge the two lists
        dataLines : List DataPoint
        dataLines = filterVApplied ++ filterVIssued


        accessor : Line.Accessor DataPoint
        accessor =
            Line.cont
                { xGroup = .groupLabel >> Just
                , xValue = .x >> toFloat
                , yValue = .y >> toFloat
                }

        sharedAttributes : List (Axis.Attribute value)
        sharedAttributes =
            [ Axis.tickSizeOuter 0
            , Axis.tickSizeInner 3
            ]


        yAxis : Line.YAxis Float
        yAxis =
            Line.axisLeft (Axis.tickCount 6 :: sharedAttributes)


        xAxisTicks : List Float
        xAxisTicks =
            getYears
                |> List.map toFloat


        xAxis : Line.XAxis Float
        xAxis =
            Line.axisBottom
                ([ Axis.ticks xAxisTicks
                , Axis.tickFormat (round >> String.fromInt)
                ]
                    ++ sharedAttributes
                )


        tooltip : Html Msg
        tooltip =
            let
                margin = { top = 10, right = 80, bottom = 20, left = 30 }

                height = 600

                xOffset =
                    margin.left + 90

                annotation =
                    model.pointAnnotation

                bottom =
                    annotation
                        |> Maybe.map Tuple.first
                        |> Maybe.map
                            (\{ yPosition } -> String.fromFloat (height - yPosition) ++ "px")
                            -- (\{ yPosition } -> String.fromFloat (height - 40) ++ "px")
                        |> Maybe.withDefault ""

                left =
                    annotation
                        |> Maybe.map Tuple.first
                        |> Maybe.map
                            (\{ xPosition } -> String.fromFloat (xPosition + xOffset) ++ "px")
                            -- (\{ xPosition } -> String.fromFloat (width - 20) ++ "px")
                        |> Maybe.withDefault ""

                display =
                    annotation
                        |> Maybe.map (always "block")
                        |> Maybe.withDefault "none"

                year =
                    annotation
                        |> Maybe.map
                            (Tuple.first
                                >> .selection
                                >> .x
                                >> String.fromFloat
                            )
                        |> Maybe.withDefault ""

                values =
                    annotation
                        |> Maybe.map
                            (Tuple.first
                                >> .selection
                                >> .y
                            )
                        |> Maybe.withDefault []
                        |> List.map
                            (\{ groupLabel, value } ->
                                Html.div []
                                    [ Html.text
                                        (Maybe.withDefault "" groupLabel
                                            ++ ": "
                                            ++ String.fromFloat value
                                        )
                                    ]
                            )

                rate = annotation
                    |> Maybe.map
                        (Tuple.first
                            >> .selection
                            >> .y
                        )
                    |> Maybe.withDefault []
                    |> List.map
                            (\{ groupLabel, value } ->
                                value
                            )
                    |> List.foldl (\x y -> x / y) 1.0
                    |> String.fromFloat
                    |> String.left 5
            in
            Html.div
                [ Html.Attributes.style "margin" "0"
                , Html.Attributes.style "padding" "5px"
                , Html.Attributes.style "font-size" "14px"
                , Html.Attributes.style "border" "1px #aaa solid"
                , Html.Attributes.style "position" "absolute"
                , Html.Attributes.style "bottom" bottom
                , Html.Attributes.style "left" left
                , Html.Attributes.style "background" "white"
                , Html.Attributes.style "opacity" "90%"
                , Html.Attributes.style "width" "130px"
                , Html.Attributes.style "display" display
                ]
                [ Html.text year
                , Html.br [] []
                , Html.div [] values
                , Html.br [] []
                , Html.text ("ratio: " ++ rate)
                ]


        -- -- Create the Line Plot
        chart : List DataPoint -> Html Msg
        chart data =
            let
                linit = Line.init
                    { margin = { top = 10, right = 80, bottom = 40, left = 60 }
                    , width = 800
                    , height = 600
                    }
                    |> Line.withColorPalette Scale.Color.tableau10
                    |> Line.withEvent (Line.hoverAll Local.Hint)
                    |> Line.withPointAnnotation model.pointAnnotation
                    |> Line.withVLineAnnotation model.vLineAnnotation
                    -- |> Line.withLabels Line.xGroupLabel
                    |> Line.withSymbols [ 
                        Symbol.circle
                        |> Symbol.withStyle [ ( "stroke", "white" ) ]
                        |> Symbol.withSize 8 ]
                    |> Line.withLineStyle [ ( "stroke-width", "2.5" ) ]
                    |> Line.withYAxis yAxis
                    |> Line.withXAxisCont xAxis
                    |> Line.render ( data, accessor )

            in
            Html.div [ Html.Attributes.style "position" "relative"
                     ]
                [ Html.map propagate linit
                , tooltip
                ]



        -- Handle country change
        handleCountryChange : String -> Msg
        handleCountryChange msg =
            msg
            |> Local.CountryChange
            |> propagate

        -- Handle consulate change
        handleConsulateChange : String -> Msg
        handleConsulateChange msg =
            msg
            |> Local.ConsulateChange
            |> propagate

    in
    div [  Html.Attributes.class "barplot-wrapper"  ]
        [ div [ Html.Attributes.style "display" "flex"
            , Html.Attributes.style "flex-direction" "column"
            , Html.Attributes.style "gap" "4px"
        ] [
            div [ Html.Attributes.style "display" "flex"
                , Html.Attributes.style "flex-direction" "column"
                , Html.Attributes.style "gap" "6px"] [

                div [ Html.Attributes.style "display" "flex", Html.Attributes.style "flex-direction" "row", Html.Attributes.style "gap" "6px" ] [
                    label [Html.Attributes.for "select_country"] [ Html.text "Choose Schengen State: " ]
                    , select
                        [ Html.Attributes.id "select_country"
                        , Html.Attributes.name "country"
                        , Html.Attributes.style "width" "10rem"
                        , onInput handleCountryChange
                        ] 
                            (List.map country_option countries)
                ]
                , div [ Html.Attributes.style "display" "flex", Html.Attributes.style "flex-direction" "row", Html.Attributes.style "gap" "6px" ] [
                    label [Html.Attributes.for "select_consulate"] [ Html.text "Choose Consulate Country: " ]
                    , select
                        [ Html.Attributes.id "select_consulate"
                        , Html.Attributes.name "consulate"
                        , Html.Attributes.style "width" "10rem"
                        , onInput handleConsulateChange
                        ] 
                            (List.map consulate_option consulate_countries)
                ]
            ]
        ] 
        , div [ Html.Attributes.style "font-size" "12px"
                , Html.Attributes.style "margin-left" "4rem"
                , Html.Attributes.style "margin-bottom" "-1rem"
                , Html.Attributes.style "margin-top" "1rem" ]
            [ Html.text "Number of Visas"]

        , chart dataLines

        , div [ Html.Attributes.style "font-size" "12px"
                , Html.Attributes.style "margin" "auto"
                , Html.Attributes.style "margin-top" "-2rem"
                , Html.Attributes.style "margin-bottom" "1rem" ] 
            [ Html.text "Years"]
        , div [ Html.Attributes.style "display" "flex"
                , Html.Attributes.style "flex-direction" "row"
                , Html.Attributes.style "gap" "4px"
                , Html.Attributes.style "margin" "auto"
        ] [
            div [ Html.Attributes.style "height" "15px"
                 , Html.Attributes.style "width" "15px"
                 , Html.Attributes.style "background-color" "#4e79a7"
                 , Html.Attributes.style "border-radius"" 50%"
                 , Html.Attributes.style "display" "inline-block"] []
            , div [] [ Html.text "applied" ]
            , div [ Html.Attributes.style "height" "15px"
                 , Html.Attributes.style "width" "15px"
                 , Html.Attributes.style "background-color" "#f28e2c"
                 , Html.Attributes.style "border-radius"" 50%"
                 , Html.Attributes.style "display" "inline-block"] []
            , div [] [ Html.text "issued" ]
        ]
        ]